#include "DistrhoUI.hpp"
#include "tablet.h"
#include <format>

START_NAMESPACE_DISTRHO

class TabUI : public UI
{
	static const uint kInitialWidth = 405;
	static const uint kInitialHeight = 256;

public:
	TabUI()
		: UI(kInitialWidth, kInitialHeight),
		fSampleRate(getSampleRate()),
		fResizable(isResizable()),
		tab(getWindow().getNativeWindowHandle())
	{

#ifdef DGL_NO_SHARED_RESOURCES
		createFontFromFile("sans", "/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf");
#else
		loadSharedResources();
#endif

		setGeometryConstraints(kInitialWidth, kInitialHeight, true);
	}

protected:

	void getTabletData()
	{
		if (!tab.initialized || !tab.GetPacket(pkt)) return;
		if (pkt == lastPkt) return;
		if (pkt.x != lastPkt.x)				setParameterValue(ktpax, pkt.x);
		if (pkt.y != lastPkt.y)				setParameterValue(ktpay, pkt.y);
		if (pkt.z != lastPkt.z)				setParameterValue(ktpaz, pkt.z);
		if (pkt.p != lastPkt.p)				setParameterValue(ktpap, pkt.p);
		if (pkt.buttons != lastPkt.buttons) setParameterValue(ktpab, pkt.buttons);
		lastPkt = pkt;
	}

	void parameterChanged(uint32_t index, float value) override
	{

		if (index != kParameterTime && index < kParameterCount) {
			switch (index)
			{
			case(ktpax): pkt.x = value;
			case(ktpay): pkt.y = value;
			case(ktpaz): pkt.z = value;
			case(ktpap): pkt.p = value;
			case(ktpab): pkt.buttons = static_cast<unsigned long>(value);
			}
			return;
		}

		getTabletData();
		repaint();
	}


	void sampleRateChanged(double newSampleRate) override
	{
		fSampleRate = newSampleRate;
		repaint();
	}

	void uiIdle() override
	{
		getTabletData();
		repaint();
	}

	bool onMouse(const MouseEvent& ev) override
	{
		getTabletData();
		repaint();
		return false; //Allow event to propagate.
	}


	void onNanoDisplay() override
	{
		fontSize(15.0f);
		textLineHeight(1.f);

		//Report tablet errors.
		if (!tab.initialized) {
			const std::string err = std::vformat("Tablet not supported:\n{}",
				std::make_format_args(tab.errormsg));
			beginPath();
			fillColor(200, 200, 200);
		//	textBox(0.f, 15.f, 250.f, err.c_str(), nullptr);
			closePath();
			return;
		}

		//Numerical feedback.
		beginPath();
		fillColor(200, 200, 200);
		textBox(0.f, 15.f, 250.f,
			std::format("x: {:.3f}\ny: {:.3f}\nz: {:.3f}\np: {:.3f}\nb: {}",
				pkt.x, pkt.y, pkt.z, pkt.p, pkt.buttons).c_str(), nullptr);
		closePath();

		//Pen position and pressure.
		drawCircle(pkt.x, pkt.y, pkt.z, pkt.p);
	}

	void drawCircle(float x, float y, float z, float p) {

		static constexpr float circleRadius = 25.f;
		x *= getWidth();
		y = (1.f - y) * getHeight();
		z = 1.f - z;

		beginPath();
		strokeColor(1.f, 1.f, 1.f, 0.5f);
		moveTo(x - z * circleRadius, y);
		lineTo(x + z * circleRadius, y);
		stroke();
		closePath();

		beginPath();
		strokeColor(1.f, 1.f, 1.f, 0.5f);
		moveTo(x, y - z * circleRadius);
		lineTo(x, y + z * circleRadius);
		stroke();
		closePath();

		beginPath();
		fillColor(1.f, 1.f, 1.f, p);
		strokeColor(255, 255, 255, 255);
		circle(x, y, circleRadius);
		fill();
		stroke();
		closePath();

		beginPath();
		strokeColor(1.f, 1.f, 1.f, z);
		circle(x, y, z * circleRadius);
		stroke();
		closePath();
	}

	void onResize(const ResizeEvent& ev) override
	{
		fScale = static_cast<float>(ev.size.getHeight()) / static_cast<float>(kInitialHeight);

		UI::onResize(ev);
	}

	void uiScaleFactorChanged(const double scaleFactor) override
	{
		fScaleFactor = scaleFactor;
	}

	// -------------------------------------------------------------------------------------------------------

private:
	// Parameters
	double fSampleRate;

	// UI stuff
	bool fResizable;
	float fScale; // our internal scaling
	double fScaleFactor; // host reported scale factor

	// Tablet context handler
	Tablet tab;
	Packet pkt = { 0 };
	Packet lastPkt = { 0 };


	/**
	   Set our UI class as non-copyable and add a leak detector just in case.
	 */
	DISTRHO_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR(TabUI)
};



UI* createUI()
{
	return new TabUI();
}


END_NAMESPACE_DISTRHO
